-- Bump
bump = require 'lib.bump'
bump.initialize(50)

entities = {}
entities.objects = {}
entities.selected = {}
entities.count = 0
local register = {}
local id = 0

-- Registers entities to be used
function entities.init()
	entities.register('box')
end

-- Loads the entity's file
function entities.register(name)
	register[name] = love.filesystem.load('entity/'..name..'.lua')
end


-- Loads a file to serve as a template for an entity
function entities.derive(name)
	return love.filesystem.load('entity/'..name..'.lua')()
end


function bump.collision(entity1,entity2, dx,dy)
	-- print(entity1.id, "collision with", entity2.id, "displacement vector:", dx, dy)
	local x1,y1 = entity1:getPos()
	local x2,y2 = entity2:getPos()
	local tx1,ty1 = entity1:getPos()
	local tx2,ty2 = entity2:getPos()
	local s1 = entity1:getSight()
	local s2 = entity2:getSight()
	
	entity1:setPos(x1+dx/2, y1+dy/2)
	entity2:setPos(x2-dx/2, y2-dy/2)
	
	-- if vector.dist(x1,y1, tx1,ty1) <= s1 then
	-- 	entity1:setTargetPos(x1,y1)
	-- end
	-- if vector.dist(x2,y2, tx2,ty2) <= s2 then
	-- 	entity2:setTargetPos(x2,y2)
	-- end
	
	if entity1:getTeam() ~= entity2:getTeam() then
		entity1:attack(entity2)
		entity2:attack(entity1)
	end
end

function bump.endCollision(entity, target)
	-- print(entity.id, "stopped colliding with", target.id)
end

function bump.getBBox(entity)
	return entity.x, entity.y, entity.w, entity.h
end

function bump.shouldCollide(entity, target)
	return true -- we could add certain conditions here - for example, make objects of the same group not collide
end


-- Creates a new entity
function entities.create(name, x,y, team)
	x,y, team = x or 0, y or 0, team or 0
	
	-- The magical code that makes the entity system work
	if register[name] then
		id = id + 1
		local entity = register[name]()
		
		entity:load()
		
		entity.type = name
		entity:setPos(x,y)
		entity:setTargetPos(x,y)
		entity:setTeam(team)
		entity:setSize(entity.image:getWidth(), entity.image:getHeight())
		entity.id = id
		
		entities.objects[id] = entity
		entities.count = entities.count + 1
		bump.add(entities.objects[id])
		
		return entities.count
	else
		print("ERROR: Entity "..name.." does not exist!")
	end
end

-- Removes an entity
function entities.destroy(id)
	if entities.objects[id] then
		if entities.objects[id].die then
			entities.objects[id]:die()
		end
		entities.count = entities.count - 1
		bump.remove(entities.objects[id])
		entities.objects[id] = nil
	end
end

-- Removes an entity
function entities.destroyAll()
	for i, entity in pairs(entities.objects) do
		entities.destroy(entity.id)
	end
end


-- Updates any applicable entity
function entities.update(dt)
	for i, entity in pairs(entities.objects) do
		if entity.update then
			-- if entity:isIdle() then
				entities.searchForTarget(entity)
			-- end
			entity:update(dt)
		end
	end
end

-- Draws any applicable entity
function entities.draw()
	for i, entity in pairs(entities.objects) do
		if entity.draw then
			entity:draw()
		end
	end
end


-- Select the entities in the selection rectangle
function entities.select(x,y, w,h)
	for i, entity in pairs(entities.objects) do
		if entity:getTeam() == playerTeam then
			if entity.select then
				if insideSelection(x,y, w,h, entity.x,entity.y, entity.w,entity.h) then
					entity:select()
				end
			end
		end
	end
end

-- Select all entities
function entities.selectAll()
	for i, entity in pairs(entities.objects) do
		if entity:getTeam() == playerTeam then
			if entity.select then
				entity:select()
			end
		end
	end
end

-- Deselect all entities
function entities.deselectAll()
	for i, entity in pairs(entities.objects) do
		if entity.select then
			entity:deselect()
		end
	end
end


-- Sets the target coordinates for a selected entity
function entities.setTargetPos(x,y)
	for i, entity in pairs(entities.objects) do
		if entity:isSelected() then
			entity:setTargetPos(x,y)
		end
	end
end

-- Searches for target and sets it as target
function entities.searchForTarget(entity)
	for i, target in pairs(entities.objects) do
		if target.health then
			if entity:getTeam() ~= target:getTeam() then
				local ex,ey = entity:getPos()
				local tx,ty = target:getPos()
				if vector.dist(ex,ey, tx,ty) <= entity:getSight() then
					-- print("Entity:"..entity.team..entity.type..entity.id.." Target:"..target.team..target.type..target.id)
					entity:setTargetEntity(target)
				end
			end
		end
	end
end


-- Draws the selection indicators around selected entities
function entities.drawSelectionIndicators()
	for i, entity in pairs(entities.objects) do
		if entity:isSelected() then
			local x,y = cam:cameraCoords(entity:getPos())
			local w,h = entity:getSize()
			
			love.graphics.setLine(1, 'smooth')
			love.graphics.setColor(255,255,255,55)
			love.graphics.rectangle('fill', x,y + h*cam.zoom/2, w*cam.zoom,h*cam.zoom/2)
			love.graphics.setColor(255,255,255,255)
			love.graphics.rectangle('line', x,y + h*cam.zoom/2, w*cam.zoom,h*cam.zoom/2)
		end
	end
end

-- Draws the health indicator above entities with health
function entities.drawHealthIndicators()
	for i, entity in pairs(entities.objects) do
		-- if entity:getTeam() == playerTeam then
			if entity.health then
				local x,y = entity:getPos()
				local w,h = entity:getSize()
				local h,m = entity:getHealth()
				
				love.graphics.setLine(1, 'smooth')
				love.graphics.setColor(255,0,0,255)
				love.graphics.rectangle('fill', x-(m/5-w)/2,y-1, entity.maxHealth/5,1)
				love.graphics.setColor(0,255,0,255)
				love.graphics.rectangle('fill', x-(m/5-w)/2,y-1, entity.health/5,1)
			end
		-- end
	end
end

-- Draws the line of sight circles for entities
function entities.drawLOS()
	for i, entity in pairs(entities.objects) do
		-- if entity:getTeam() == playerTeam then
			if entity.sight then
				local x,y = entity:getPos()
				local s = entity:getSight()
				
				love.graphics.setLine(1, 'smooth')
				love.graphics.setColor(255,255,255,25)
				love.graphics.circle('line', x,y, s, 16)
			end
		-- end
	end
end
